/* Frustrum Culling Demo

   (Demonstrates checking if shapes are in view)

   CFrustrum.cpp


  KOS version:   Heinrich Tillack    tillack@web.de
  original author : DigiBen	     digiben@gametutorials.com

   Thanks to Jim Ursetto`s idea for glGetFloatv()

*/


#include <kos.h>
#include <GL/gl.h>
#include <GL/glu.h>


#include "CFrustrum.h"



// We create an enum of the sides so we don't have to call each side 0 or 1.
// This way it makes it more understandable and readable when dealing with frustum sides.
enum FrustumSide
{
    RIGHT	= 0,		// The RIGHT side of the frustum
    LEFT	= 1,		// The LEFT	 side of the frustum
    BOTTOM	= 2,		// The BOTTOM side of the frustum
    TOP	= 3,		// The TOP side of the frustum
    BACK	= 4,		// The BACK	side of the frustum
    FRONT	= 5			// The FRONT side of the frustum
};

// Like above, instead of saying a number for the ABC and D of the plane, we
// want to be more descriptive.
enum PlaneData
{
    A = 0,				// The X value of the plane's normal
    B = 1,				// The Y value of the plane's normal
    C = 2,				// The Z value of the plane's normal
    D = 3				// The distance the plane is from the origin
};


///////////////////////////////// NORMALIZE PLANE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
/////
/////	This normalizes a plane (A side) from a given frustum.
/////
///////////////////////////////// NORMALIZE PLANE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*

inline void NormalizePlane(float frustum[6][4], int side)
{
    // Here we calculate the magnitude of the normal to the plane (point A B C)
    // Remember that (A, B, C) is that same thing as the normal's (X, Y, Z).
    // To calculate magnitude you use the equation:  magnitude = sqrt( x^2 + y^2 + z^2)
    float magnitude = (float)fsqrt( frustum[side][A] * frustum[side][A] +
                                    frustum[side][B] * frustum[side][B] +
                                    frustum[side][C] * frustum[side][C] );

    // Then we divide the plane's values by it's magnitude.
    // This makes it easier to work with.
    frustum[side][A] /= magnitude;
    frustum[side][B] /= magnitude;
    frustum[side][C] /= magnitude;
    frustum[side][D] /= magnitude;
}


///////////////////////////////// CALCULATE FRUSTUM \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
/////
/////	This extracts our frustum from the projection and modelview matrix.
/////
///////////////////////////////// CALCULATE FRUSTUM \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*





matrix_t modl __attribute__((aligned(32)));
matrix_t proj __attribute__((aligned(32)));
matrix_t clip __attribute__((aligned(32)));

matrix_t temp __attribute__((aligned(32)));


void CFrustum::CalculateFrustum()
{


    //	float   proj[16];								// This will hold our projection matrix
    //	float   modl[16];								// This will hold our modelview matrix
    //	float   clip[16];								// This will hold the clipping planes

    // glGetFloatv() is used to extract information about our OpenGL world.
    // Below, we pass in GL_PROJECTION_MATRIX to abstract our projection matrix.
    // It then stores the matrix into an array of [16].

    	glGetFloatv( GL_PROJECTION, (GLfloat *)&proj );

//     glLoadIdentity();
//     glKosMatrixApply(GL_PROJECTION);
//     mat_store(&proj );


    // By passing in GL_MODELVIEW_MATRIX, we can abstract our model view matrix.
    // This also stores it in an array of [16].

   	glGetFloatv( GL_MODELVIEW , (GLfloat *)&modl );
//     glLoadIdentity();
//     glKosMatrixApply(GL_MODELVIEW);
//     mat_store(&modl);

    // Now that we have our modelview and projection matrix, if we combine these 2 matrices,
    // it will give us our clipping planes.
    //To combine 2 matrices, we multiply them with the SH4.

    // clip=modl*proj


    //XXX store the internal SH4 matrix
    mat_store(&temp);

    // 	 mat_load(&proj);
    // 	 mat_apply(&modl);

    mat_load(&modl);
    mat_apply(&proj);

    mat_store(&clip);

    //XXX do we need this restore function?
    mat_load(&temp);


    //	clip[ 0] = modl[ 0] * proj[ 0] + modl[ 1] * proj[ 4] + modl[ 2] * proj[ 8] + modl[ 3] * proj[12];
    //	clip[ 1] = modl[ 0] * proj[ 1] + modl[ 1] * proj[ 5] + modl[ 2] * proj[ 9] + modl[ 3] * proj[13];
    //	clip[ 2] = modl[ 0] * proj[ 2] + modl[ 1] * proj[ 6] + modl[ 2] * proj[10] + modl[ 3] * proj[14];
    //	clip[ 3] = modl[ 0] * proj[ 3] + modl[ 1] * proj[ 7] + modl[ 2] * proj[11] + modl[ 3] * proj[15];
    //
    //	clip[ 4] = modl[ 4] * proj[ 0] + modl[ 5] * proj[ 4] + modl[ 6] * proj[ 8] + modl[ 7] * proj[12];
    //	clip[ 5] = modl[ 4] * proj[ 1] + modl[ 5] * proj[ 5] + modl[ 6] * proj[ 9] + modl[ 7] * proj[13];
    //	clip[ 6] = modl[ 4] * proj[ 2] + modl[ 5] * proj[ 6] + modl[ 6] * proj[10] + modl[ 7] * proj[14];
    //	clip[ 7] = modl[ 4] * proj[ 3] + modl[ 5] * proj[ 7] + modl[ 6] * proj[11] + modl[ 7] * proj[15];
    //
    //	clip[ 8] = modl[ 8] * proj[ 0] + modl[ 9] * proj[ 4] + modl[10] * proj[ 8] + modl[11] * proj[12];
    //	clip[ 9] = modl[ 8] * proj[ 1] + modl[ 9] * proj[ 5] + modl[10] * proj[ 9] + modl[11] * proj[13];
    //	clip[10] = modl[ 8] * proj[ 2] + modl[ 9] * proj[ 6] + modl[10] * proj[10] + modl[11] * proj[14];
    //	clip[11] = modl[ 8] * proj[ 3] + modl[ 9] * proj[ 7] + modl[10] * proj[11] + modl[11] * proj[15];
    //
    //	clip[12] = modl[12] * proj[ 0] + modl[13] * proj[ 4] + modl[14] * proj[ 8] + modl[15] * proj[12];
    //	clip[13] = modl[12] * proj[ 1] + modl[13] * proj[ 5] + modl[14] * proj[ 9] + modl[15] * proj[13];
    //	clip[14] = modl[12] * proj[ 2] + modl[13] * proj[ 6] + modl[14] * proj[10] + modl[15] * proj[14];
    //	clip[15] = modl[12] * proj[ 3] + modl[13] * proj[ 7] + modl[14] * proj[11] + modl[15] * proj[15];



    // Now we actually want to get the sides of the frustum.  To do this we take
    // the clipping planes we received above and extract the sides from them.

    // This will extract the RIGHT side of the frustum

    //	m_Frustum[RIGHT][A] = clip[ 3] - clip[ 0];
    //	m_Frustum[RIGHT][B] = clip[ 7] - clip[ 4];
    //	m_Frustum[RIGHT][C] = clip[11] - clip[ 8];
    //	m_Frustum[RIGHT][D] = clip[15] - clip[12];

    m_Frustum[RIGHT][A] = clip[0][3] - clip[0][0];
    m_Frustum[RIGHT][B] = clip[1][3] - clip[1][0];
    m_Frustum[RIGHT][C] = clip[2][3] - clip[2][0];
    m_Frustum[RIGHT][D] = clip[3][3] - clip[3][0];

    // Now that we have a normal (A,B,C) and a distance (D) to the plane,
    // we want to normalize that normal and distance.

    // Normalize the RIGHT side
    NormalizePlane(m_Frustum, RIGHT);

    // This will extract the LEFT side of the frustum
    //	m_Frustum[LEFT][A] = clip[ 3] + clip[ 0];
    //	m_Frustum[LEFT][B] = clip[ 7] + clip[ 4];
    //	m_Frustum[LEFT][C] = clip[11] + clip[ 8];
    //	m_Frustum[LEFT][D] = clip[15] + clip[12];

    m_Frustum[LEFT][A] = clip[0][3] + clip[0][0];
    m_Frustum[LEFT][B] = clip[1][3] + clip[1][0];
    m_Frustum[LEFT][C] = clip[2][3] + clip[2][0];
    m_Frustum[LEFT][D] = clip[3][3] + clip[3][0];

    // Normalize the LEFT side
    NormalizePlane(m_Frustum, LEFT);

    // This will extract the BOTTOM side of the frustum
    //	m_Frustum[BOTTOM][A] = clip[ 3] + clip[ 1];
    //	m_Frustum[BOTTOM][B] = clip[ 7] + clip[ 5];
    //	m_Frustum[BOTTOM][C] = clip[11] + clip[ 9];
    //	m_Frustum[BOTTOM][D] = clip[15] + clip[13];
    //
    m_Frustum[BOTTOM][A] = clip[0][3] + clip[0][1];
    m_Frustum[BOTTOM][B] = clip[1][3] + clip[1][1];
    m_Frustum[BOTTOM][C] = clip[2][3] + clip[2][1];
    m_Frustum[BOTTOM][D] = clip[3][3] + clip[3][1];

    // Normalize the BOTTOM side
    NormalizePlane(m_Frustum, BOTTOM);

    // This will extract the TOP side of the frustum
    //	m_Frustum[TOP][A] = clip[ 3] - clip[ 1];
    //	m_Frustum[TOP][B] = clip[ 7] - clip[ 5];
    //	m_Frustum[TOP][C] = clip[11] - clip[ 9];
    //	m_Frustum[TOP][D] = clip[15] - clip[13];

    m_Frustum[TOP][A] = clip[0][3] - clip[0][1];
    m_Frustum[TOP][B] = clip[1][3] - clip[1][1];
    m_Frustum[TOP][C] = clip[2][3] - clip[2][1];
    m_Frustum[TOP][D] = clip[3][3] - clip[3][1];

    // Normalize the TOP side
    NormalizePlane(m_Frustum, TOP);

    // This will extract the BACK side of the frustum
    //	m_Frustum[BACK][A] = clip[ 3] - clip[ 2];
    //	m_Frustum[BACK][B] = clip[ 7] - clip[ 6];
    //	m_Frustum[BACK][C] = clip[11] - clip[10];
    //	m_Frustum[BACK][D] = clip[15] - clip[14];
    //
    m_Frustum[BACK][A] = clip[0][3] - clip[0][2];
    m_Frustum[BACK][B] = clip[1][3] - clip[1][2];
    m_Frustum[BACK][C] = clip[2][3] - clip[2][2];
    m_Frustum[BACK][D] = clip[3][3] - clip[3][2];

    // Normalize the BACK side
    NormalizePlane(m_Frustum, BACK);

    // This will extract the FRONT side of the frustum
    //	m_Frustum[FRONT][A] = clip[ 3] + clip[ 2];
    //	m_Frustum[FRONT][B] = clip[ 7] + clip[ 6];
    //	m_Frustum[FRONT][C] = clip[11] + clip[10];
    //	m_Frustum[FRONT][D] = clip[15] + clip[14];
    //
    m_Frustum[FRONT][A] = clip[0][3] + clip[0][2];
    m_Frustum[FRONT][B] = clip[1][3] + clip[1][2];
    m_Frustum[FRONT][C] = clip[2][3] + clip[2][2];
    m_Frustum[FRONT][D] = clip[3][3] + clip[3][2];

    // Normalize the FRONT side
    NormalizePlane(m_Frustum, FRONT);
}

// The code below will allow us to make checks within the frustum.  For example,
// if we want to see if a point, a sphere, or a cube lies inside of the frustum.
// Because all of our planes point INWARDS (The normals are all pointing inside the frustum)
// we then can assume that if a point is in FRONT of all of the planes, it's inside.

///////////////////////////////// POINT IN FRUSTUM \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
/////
/////	This determines if a point is inside of the frustum
/////
///////////////////////////////// POINT IN FRUSTUM \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*

bool CFrustum::PointInFrustum( float x, float y, float z )
{
    // If you remember the plane equation (A*x + B*y + C*z + D = 0), then the rest
    // of this code should be quite obvious and easy to figure out yourself.
    // In case don't know the plane equation, it might be a good idea to look
    // at our Plane Collision tutorial at www.GameTutorials.com in OpenGL Tutorials.
    // I will briefly go over it here.  (A,B,C) is the (X,Y,Z) of the normal to the plane.
    // They are the same thing... but just called ABC because you don't want to say:
    // (x*x + y*y + z*z + d = 0).  That would be wrong, so they substitute them.
    // the (x, y, z) in the equation is the point that you are testing.  The D is
    // The distance the plane is from the origin.  The equation ends with "= 0" because
    // that is true when the point (x, y, z) is ON the plane.  When the point is NOT on
    // the plane, it is either a negative number (the point is behind the plane) or a
    // positive number (the point is in front of the plane).  We want to check if the point
    // is in front of the plane, so all we have to do is go through each point and make
    // sure the plane equation goes out to a positive number on each side of the frustum.
    // The result (be it positive or negative) is the distance the point is front the plane.

    // Go through all the sides of the frustum
    for(int i = 0; i < 6; i++ )
    {
        // Calculate the plane equation and check if the point is behind a side of the frustum
        if(m_Frustum[i][A] * x + m_Frustum[i][B] * y + m_Frustum[i][C] * z + m_Frustum[i][D] <= 0)
        {
            // The point was behind a side, so it ISN'T in the frustum
            return false;
        }
    }

    // The point was inside of the frustum (In front of ALL the sides of the frustum)
    return true;
}


///////////////////////////////// SPHERE IN FRUSTUM \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
/////
/////	This determines if a sphere is inside of our frustum by it's center and radius.
/////
///////////////////////////////// SPHERE IN FRUSTUM \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*

int CFrustum::SphereInFrustum( float x, float y, float z, float radius )
{

    int p;
    int c = 0;
    float d;

   #define frustum m_Frustum

    for( p = 0; p < 6; p++ )
    {
        d = frustum[p][A] * x + frustum[p][B] * y + frustum[p][C] * z + frustum[p][D];
        if( d <= -radius )
            return FULL_OUTSIDE;
        if( d > radius )
            c++;
    }

    //return  PARTIAL_INSIDE;
    return (c == 6) ? FULL_INSIDE :	 PARTIAL_INSIDE;

   #undef frustum
}


///////////////////////////////// CUBE IN FRUSTUM \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
/////
/////	This determines if a cube is in or around our frustum by it's center and 1/2 it's length
/////
///////////////////////////////// CUBE IN FRUSTUM \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*

bool CFrustum::CubeInFrustum( float x, float y, float z, float size )
{
    // This test is a bit more work, but not too much more complicated.
    // Basically, what is going on is, that we are given the center of the cube,
    // and half the length.  Think of it like a radius.  Then we checking each point
    // in the cube and seeing if it is inside the frustum.  If a point is found in front
    // of a side, then we skip to the next side.  If we get to a plane that does NOT have
    // a point in front of it, then it will return false.

    // *Note* - This will sometimes say that a cube is inside the frustum when it isn't.
    // This happens when all the corners of the bounding box are not behind any one plane.
    // This is rare and shouldn't effect the overall rendering speed.

    for(int i = 0; i < 6; i++ )
    {
        if(m_Frustum[i][A] * (x - size) + m_Frustum[i][B] * (y - size) + m_Frustum[i][C] * (z - size) + m_Frustum[i][D] > 0)
            continue;
        if(m_Frustum[i][A] * (x + size) + m_Frustum[i][B] * (y - size) + m_Frustum[i][C] * (z - size) + m_Frustum[i][D] > 0)
            continue;
        if(m_Frustum[i][A] * (x - size) + m_Frustum[i][B] * (y + size) + m_Frustum[i][C] * (z - size) + m_Frustum[i][D] > 0)
            continue;
        if(m_Frustum[i][A] * (x + size) + m_Frustum[i][B] * (y + size) + m_Frustum[i][C] * (z - size) + m_Frustum[i][D] > 0)
            continue;
        if(m_Frustum[i][A] * (x - size) + m_Frustum[i][B] * (y - size) + m_Frustum[i][C] * (z + size) + m_Frustum[i][D] > 0)
            continue;
        if(m_Frustum[i][A] * (x + size) + m_Frustum[i][B] * (y - size) + m_Frustum[i][C] * (z + size) + m_Frustum[i][D] > 0)
            continue;
        if(m_Frustum[i][A] * (x - size) + m_Frustum[i][B] * (y + size) + m_Frustum[i][C] * (z + size) + m_Frustum[i][D] > 0)
            continue;
        if(m_Frustum[i][A] * (x + size) + m_Frustum[i][B] * (y + size) + m_Frustum[i][C] * (z + size) + m_Frustum[i][D] > 0)
            continue;

        // If we get here, it isn't in the frustum
        return false;
    }

    return true;
}




int CFrustum::CubeLocation(float fPoint[], float fSize)
{
    ////////////////////////////////////////////////////////////////////////
    // Return FULL_OUTSIDE, PARTIAL_INSIDE or FULL_INSIDE. The parameter
    // fSize has to be the width of the cube divided by two
    ////////////////////////////////////////////////////////////////////////

    unsigned int iCurPlane;
    int iCount1;
    int iCount2 = 0;

	#define m_fFrustum m_Frustum

    for (iCurPlane = 0; iCurPlane<6; iCurPlane++)
    {
        iCount1 = 0;

        if (m_fFrustum[iCurPlane][0] * (fPoint[0] - fSize) + m_fFrustum[iCurPlane][1] *
                (fPoint[1] - fSize) + m_fFrustum[iCurPlane][2] * (fPoint[2] - fSize) +
                m_fFrustum[iCurPlane][3] > 0)
            iCount1++;
        if (m_fFrustum[iCurPlane][0] * (fPoint[0] + fSize) + m_fFrustum[iCurPlane][1] *
                (fPoint[1] - fSize) + m_fFrustum[iCurPlane][2] * (fPoint[2] - fSize) +
                m_fFrustum[iCurPlane][3] > 0)
            iCount1++;
        if (m_fFrustum[iCurPlane][0] * (fPoint[0] - fSize) + m_fFrustum[iCurPlane][1] *
                (fPoint[1] + fSize) + m_fFrustum[iCurPlane][2] * (fPoint[2] - fSize) +
                m_fFrustum[iCurPlane][3] > 0)
            iCount1++;
        if (m_fFrustum[iCurPlane][0] * (fPoint[0] + fSize) + m_fFrustum[iCurPlane][1] *
                (fPoint[1] + fSize) + m_fFrustum[iCurPlane][2] * (fPoint[2] - fSize) +
                m_fFrustum[iCurPlane][3] > 0)
            iCount1++;
        if (m_fFrustum[iCurPlane][0] * (fPoint[0] - fSize) + m_fFrustum[iCurPlane][1] *
                (fPoint[1] - fSize) + m_fFrustum[iCurPlane][2] * (fPoint[2] + fSize) +
                m_fFrustum[iCurPlane][3] > 0)
            iCount1++;
        if (m_fFrustum[iCurPlane][0] * (fPoint[0] + fSize) + m_fFrustum[iCurPlane][1] *
                (fPoint[1] - fSize) + m_fFrustum[iCurPlane][2] * (fPoint[2] + fSize) +
                m_fFrustum[iCurPlane][3] > 0)
            iCount1++;
        if (m_fFrustum[iCurPlane][0] * (fPoint[0] - fSize) + m_fFrustum[iCurPlane][1] *
                (fPoint[1] + fSize) + m_fFrustum[iCurPlane][2] * (fPoint[2] + fSize) +
                m_fFrustum[iCurPlane][3] > 0)
            iCount1++;
        if (m_fFrustum[iCurPlane][0] * (fPoint[0] + fSize) + m_fFrustum[iCurPlane][1] *
                (fPoint[1] + fSize) + m_fFrustum[iCurPlane][2] * (fPoint[2] + fSize) +
                m_fFrustum[iCurPlane][3] > 0)
            iCount1++;

        if (iCount1 == 0)
            return FULL_OUTSIDE;

        if (iCount1 == 8)
            iCount2++;
    }

    return (iCount2 == 6) ? FULL_INSIDE : PARTIAL_INSIDE;
}


// int CFrustum::int AABBvsFrustum(AABB *b, FRUSTUM *f)
// {
//  float m, n; int i, result = INSIDE;

//  for (i = 0; i < 6; i++) { PLANE *p = f->plane + i;

//      m = (b->mx * p->a) + (b->my * p->b) + (b->mz * p->c) + p->d;
//      n = (b->dx * fabs(p->a)) + (b->dy * fabs(p->b)) + (b->dz * fabs(p->c));

//      if (m + n < 0) return OUTSIDE;
//      if (m - n < 0) result = INTERSECT;

//      } return result;
// }
